# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .stream_io import FileStream, GZipStream, LZ4Stream
from .stream_io import FastWARCError, StreamError
from .warc import ArchiveIterator, WarcRecord, WarcRecordType

# Exposing symbols for legacy compatibility, please prefer explicit imports from submodules

__all__ = [
    "FileStream",
    "GZipStream",
    "LZ4Stream",
    "FastWARCError",
    "StreamError",
    "ArchiveIterator",
    "WarcRecord",
    "WarcRecordType"
]



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

from libcpp.string cimport string
from resiliparse_inc.string_view cimport string_view
from resiliparse_inc.cctype cimport tolower


cdef extern from * nogil:
    """
    #include <cctype>

    /**
     * Strip leading white space from a C string.
     */
    inline size_t lstrip_c_str(const char** s_ptr, size_t l) {
        const char* end = *s_ptr + l;
        while (*s_ptr < end && std::isspace((*s_ptr)[0])) {
            ++(*s_ptr);
        }
        return end - *s_ptr;
    }

    /**
     * Strip trailing white space from a C string.
     */
    inline size_t rstrip_c_str(const char** s_ptr, size_t l) {
        const char* end = *s_ptr + l;
        while (end > *s_ptr && std::isspace((end - 1)[0])) {
            --end;
        }
        return end - *s_ptr;
    }

    /**
     * Strip leading and trailing white space from a C string.
     */
    inline size_t strip_c_str(const char** s_ptr, size_t l) {
        return rstrip_c_str(s_ptr, lstrip_c_str(s_ptr, l));
    }

    /**
     * Strip trailing white space from a C++ string.
     */
    inline std::string rstrip_str(std::string&& s) {
        const char* start = s.data();
        size_t l = rstrip_c_str(&start, s.size());
        if (l != s.size()) {
            s.resize(l);
        }
        return s;
    }

    /**
     * Strip leading and trailing white space from a C++ string.
     */
    inline std::string strip_str(std::string&& s) {
        const char* start = s.data();
        size_t l = strip_c_str(&start, s.size());
        if (l != s.size()) {
            memmove(s.data(), start, l);
            s.resize(l);
        }
        return s;
    }
    """

    cdef size_t rstrip_c_str(const char** s_ptr, size_t l)
    cdef size_t lstrip_c_str(const char** s_ptr, size_t l)
    cdef size_t strip_c_str(const char** s_ptr, size_t l)

    cdef string rstrip_str(string s)
    cdef string strip_str(string s)


cdef inline string_view strip_sv(string_view s) noexcept nogil:
    """Strip leading and trailing white space from a C++ string_view."""
    cdef const char* start = s.data()
    cdef size_t l = strip_c_str(&start, s.size())
    if start != s.data():
        s.remove_prefix(start - s.data())
    if l != s.size():
        s.remove_suffix(s.size() - l)
    return s


cdef inline string str_to_lower(string s) noexcept nogil:
    """Convert a C++ string to lower-case characters."""
    cdef size_t i
    for i in range(s.size()):
        s[i] = tolower(s[i])
    return s



from datetime import datetime
from typing import (
    Union,
    Optional,
    Iterator,
    Dict,
    Tuple,
    Literal,
    Callable,
    Iterable,
    ValuesView,
    KeysView,
    BinaryIO,
)
from enum import IntFlag

from .stream_io import BufferedReader, IOStream, _GenericIOStream


class WarcRecordType(IntFlag):
    warcinfo = 2
    response = 4
    resource = 8
    request = 16
    metadata = 32
    revisit = 64
    conversion = 128
    continuation = 256
    unknown = 512
    any_type = 65535
    no_type = 0


warcinfo = WarcRecordType.warcinfo
response = WarcRecordType.response
resource = WarcRecordType.resource
request = WarcRecordType.request
metadata = WarcRecordType.metadata
revisit = WarcRecordType.revisit
conversion = WarcRecordType.conversion
continuation = WarcRecordType.continuation
unknown = WarcRecordType.unknown
no_type = WarcRecordType.no_type
any_type = WarcRecordType.any_type


class WarcHeaderMap:
    reason_phrase: Optional[str]
    status_code: Optional[str]
    status_line: str

    def append(self, key: str, value: str) -> None: ...
    def asdict(self) -> Dict[str, str]: ...
    def astuples(self) -> Tuple[Tuple[str, str], ...]: ...
    def clear(self) -> None: ...
    def get(self, key: str, default: Optional[str] = None) -> Optional[str]: ...
    def items(self) -> Iterator[Tuple[str, str]]: ...
    def keys(self) -> KeysView[str]: ...
    def values(self) -> ValuesView[str]: ...
    def write(self, stream: IOStream) -> None: ...
    def __getitem__(self, item: str) -> str: ...
    def __iter__(self) -> Iterator[Tuple[str, str]]: ...
    def __len__(self) -> int: ...
    def __setitem__(self, key: str, value: str) -> None: ...
    def __contains__(self, item: str) -> bool: ...


class WarcRecord:
    record_id: str
    record_type: WarcRecordType
    content_length: int
    record_date: Optional[datetime]
    headers: WarcHeaderMap
    is_http: bool
    is_http_parsed: bool
    http_headers: Optional[WarcHeaderMap]
    http_content_type: Optional[str]
    http_charset: Optional[str]
    http_date: Optional[datetime]
    http_last_modified: Optional[datetime]
    reader: BufferedReader
    stream_pos: int

    def init_headers(
        self, content_length: int = 0, record_type: WarcRecordType = no_type, record_urn: Optional[bytes] = None
    ) -> None: ...
    def freeze(self) -> bool: ...
    def set_bytes_content(self, content: bytes) -> None: ...
    def parse_http(self, strict_mode: bool = True, auto_decode: str = "none") -> None: ...
    def verify_block_digest(self, consume: bool = False) -> bool: ...
    def verify_payload_digest(self, consume: bool = False) -> bool: ...
    def write(
        self,
        stream: Union[IOStream, BinaryIO, _GenericIOStream],
        checksum_data: bool = False,
        payload_digest: Optional[bytes] = None,
        chunk_size: int = 16384
    ) -> int: ...



class ArchiveIterator(Iterable[WarcRecord]):
    def __init__(
        self,
        stream: Union[IOStream, BinaryIO, _GenericIOStream],
        record_types: WarcRecordType = any_type,
        parse_http: bool = True,
        min_content_length: int = -1,
        max_content_length: int = -1,
        func_filter: Optional[Callable[[WarcRecord], bool]] = None,
        verify_digests: bool = False,
        strict_mode: bool = True,
        auto_decode: Literal["none", "content", "transfer", "all"] = "none",
    ) -> None: ...
    def __iter__(self) -> Iterator[WarcRecord]: ...
    def __next__(self) -> WarcRecord: ...



from types import TracebackType
from typing import ContextManager, Optional, Type, Union, BinaryIO, Protocol

class _GenericIOStream(Protocol):
    def write(self, data: bytes) -> int: ...
    def flush(self) -> None: ...
    def read(self, size: int) -> bytes: ...
    def seek(self, offset: int) -> int: ...
    def close(self) -> None: ...
    def tell(self) -> int: ...


class IOStream(ContextManager[IOStream]):
    def read(self, size: int) -> bytes: ...
    def write(self, data: bytes) -> int: ...
    def close(self) -> None: ...
    def flush(self) -> None: ...
    def seek(self, offset: int) -> None: ...
    def tell(self) -> int: ...
    def __enter__(self) -> IOStream: ...
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> None: ...


class BufferedReader:
    def __init__(
        self, stream: Union[IOStream, BinaryIO, _GenericIOStream], buf_size: int = 65536, negotiate_stream: bool = True
    ) -> None: ...
    def close(self) -> None: ...
    def consume(self, size: int = -1) -> int: ...
    def read(self, size: int = -1) -> bytes: ...
    def readline(self, crlf: bool = True, max_line_len: int = 8192) -> bytes: ...
    def tell(self) -> int: ...


class BytesIOStream(IOStream):
    def __init__(self, initial_data: Union[bytes, None] = None) -> None: ...
    def getvalue(self) -> bytes: ...


class FileStream(IOStream):
    def __init__(self, filename: str, mode: str = "rb") -> None: ...


class CompressingStream(IOStream):
    def begin_member(self) -> int: ...
    def end_member(self) -> int: ...


class BrotliStream(CompressingStream):
    def __init__(
        self, raw_stream: Union[IOStream, BinaryIO, _GenericIOStream], quality: int = 11, lgwin: int = 22, lgblock: int = 0
    ) -> None: ...


class GZipStream(CompressingStream):
    def __init__(
        self, raw_stream: Union[IOStream, BinaryIO, _GenericIOStream], compression_level: int = 9, zlib: bool = False
    ) -> None: ...


class LZ4Stream(CompressingStream):
    def __init__(
        self,
        raw_stream: Union[IOStream, BinaryIO, _GenericIOStream],
        compression_level: int = 12,
        favor_dec_speed: bool = True,
    ) -> None: ...
    def prepopulate(self, initial_data: bytes) -> None: ...


class PythonIOStreamAdapter(IOStream):
    def __init__(self, py_stream: _GenericIOStream) -> None: ...


class FastWARCError(Exception):
    pass


class ReaderStaleError(FastWARCError):
    pass


class StreamError(FastWARCError):
    pass



from enum import IntFlag
from typing import Union, Iterator, Tuple, Protocol

from .stream_io import IOStream, _GenericIOStream
from .warc import WarcRecord


class CompressionAlg(IntFlag):
    gzip = 0
    lz4 = 1
    uncompressed = 2
    auto = 3


def detect_compression_algorithm(file: str) -> CompressionAlg: ...


def wrap_warc_stream(
    file: Union[str, IOStream, _GenericIOStream],
    mode: str,
    comp_alg: CompressionAlg = CompressionAlg.auto,
    **comp_args
) -> IOStream: ...


def recompress_warc_interactive(
    warc_in: Union[str, IOStream, _GenericIOStream],
    warc_out: Union[str, IOStream, _GenericIOStream],
    comp_alg_in: CompressionAlg = CompressionAlg.auto,
    comp_alg_out: CompressionAlg = CompressionAlg.auto,
    **comp_args
) -> Iterator[Tuple[WarcRecord, int]]: ...


def recompress_warc(
    warc_in: Union[str, IOStream, _GenericIOStream],
    warc_out: Union[str, IOStream, _GenericIOStream],
    comp_alg_in: CompressionAlg = CompressionAlg.auto,
    comp_alg_out: CompressionAlg = CompressionAlg.auto,
    **comp_args
) -> Iterator[Tuple[WarcRecord, int]]: ...


def verify_digests(
    warc_in: Union[str, IOStream],
    verify_payloads: bool = False,
    comp_alg: CompressionAlg = CompressionAlg.auto,
) -> bool: ...



# Copyright 2023 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from Cython.Build import cythonize
from Cython.Distutils.build_ext import new_build_ext as build_ext
import distutils.ccompiler
from distutils.dir_util import copy_tree
from setuptools import Extension, setup

TRACE = bool(int(os.getenv('TRACE', 0)))
DEBUG = bool(int(os.getenv('DEBUG', 0))) or TRACE
ASAN = bool(int(os.getenv('ASAN', 0)))

ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
CXX = distutils.ccompiler.get_default_compiler()


def get_cpp_args():
    cpp_args = {}

    if TRACE:
        cpp_args.update(dict(define_macros=[('CYTHON_TRACE_NOGIL', '1')]))

    if CXX == 'unix':
        cpp_args.update(dict(
            extra_compile_args=['-std=c++17',
                                f'-O{0 if DEBUG else 3}',
                                '-Wall',
                                '-Wno-deprecated-declarations',
                                '-Wno-unreachable-code',
                                '-Wno-unused-function'],
            extra_link_args=['-std=c++17']
        ))
        if DEBUG:
            cpp_args['extra_compile_args'].append('-Werror')
        if ASAN:
            cpp_args['extra_compile_args'].append('-fsanitize=address')
            cpp_args['extra_link_args'].append('-fsanitize=address')

    elif CXX == 'msvc':
        cpp_args.update(dict(
            extra_compile_args=['/std:c++latest', '/W3', '/O2'],
            extra_link_args=[]
        ))
        if DEBUG:
            cpp_args['extra_compile_args'].append('/WX')
            cpp_args['extra_link_args'].append('/WX')

    return cpp_args


def get_cython_args():
    return dict(
        annotate=DEBUG,
        gdb_debug=DEBUG,
        nthreads=os.cpu_count(),
        compiler_directives=dict(
            language_level='3',
            linetrace=TRACE,
            initializedcheck=DEBUG,
            boundscheck=DEBUG,
            cdivision=True
        )
    )


def get_ext_modules():
    cpp_args = get_cpp_args()

    fastwarc_extensions = [
        Extension('fastwarc.warc', sources=['fastwarc/warc.pyx'], **cpp_args),
        Extension('fastwarc.stream_io', sources=['fastwarc/stream_io.pyx'],
                  libraries=['zlib' if CXX == 'msvc' else 'z', 'lz4'], **cpp_args),
        Extension('fastwarc.tools', sources=['fastwarc/tools.pyx'], **cpp_args)
    ]

    return cythonize(fastwarc_extensions, **get_cython_args())


# Copy Resiliparse header files
if os.path.isdir(os.path.join(ROOT_DIR, '..', 'resiliparse', 'resiliparse_inc')):
    copy_tree(os.path.join(ROOT_DIR, '..', 'resiliparse', 'resiliparse_inc'),
              os.path.join(ROOT_DIR, 'resiliparse_inc'), update=1)
    copy_tree(os.path.join(ROOT_DIR, '..', 'resiliparse', 'resiliparse_common'),
              os.path.join(ROOT_DIR, 'resiliparse_common'), update=1)

setup(
    ext_modules=get_ext_modules(),
    cmdclass=dict(build_ext=build_ext),
    exclude_package_data={
        '': [] if 'sdist' in sys.argv else ['*.pxd', '*.pxi', '*.pyx', '*.h', '*.cpp']
    }
)
